@@ -370,18 +370,6 @@ spi_dec_x11_mouse_check (SpiDEController *controller,
                  &root_return, &child_return,
                  x, y,
                  &win_x_return, &win_y_return, &mask_return);
-  /*
-   * Since many clients grab the pointer, and X goes an automatic
-   * pointer grab on mouse-down, we often must detect mouse button events
-   * by polling rather than via a button grab.
-   * The while loop (rather than if) is used since it's possible that
-   * multiple buttons have changed state since we last checked.
-   */
-  if (mask_return != mouse_mask_state)
-    {
-      while (spi_dec_button_update_and_emit (controller, mask_return))
-        ;
-    }
 
   if (*x != last_mouse_pos->x || *y != last_mouse_pos->y)
     {
@@ -400,6 +388,19 @@ spi_dec_x11_mouse_check (SpiDEController *controller,
       *moved = False;
     }
 
+  /*
+   * Since many clients grab the pointer, and X goes an automatic
+   * pointer grab on mouse-down, we often must detect mouse button events
+   * by polling rather than via a button grab.
+   * The while loop (rather than if) is used since it's possible that
+   * multiple buttons have changed state since we last checked.
+   */
+  if (mask_return != mouse_mask_state)
+    {
+      while (spi_dec_button_update_and_emit (controller, mask_return))
+        ;
+    }
+
   return mask_return;
 }
 
@@ -720,16 +721,75 @@ spi_controller_register_with_devices (SpiDEController *controller)
   x_default_error_handler = XSetErrorHandler (_spi_controller_device_error_handler);
 }
 
+/* Compute effective XKB level from modifier state:
+ * bit0: Shift, bit1: ISO_Level3_Shift (AltGr), bit2: ISO_Level5_Shift.
+ * Lock and NumLock do not affect level.
+ */
+static unsigned int
+get_level_from_state (Display *dpy, unsigned int state)
+{
+  unsigned int level = 0;
+  unsigned int altgr_mask = XkbKeysymToModifiers (dpy, XK_ISO_Level3_Shift);
+  unsigned int level5_mask = XkbKeysymToModifiers (dpy, XK_ISO_Level5_Shift);
+
+  if (state & ShiftMask)
+    level |= 1;
+  if (altgr_mask && (state & altgr_mask))
+    level |= 2;
+  if (level5_mask && (state & level5_mask))
+    level |= 4;
+
+  return level;
+}
+
+/* Resolve keysym using explicit current group (st.group) and effective level.
+ * Fallback: scan levels in current group, then group 0 for specials.
+ */
+static KeySym
+resolve_keysym_with_xkb (Display *dpy, KeyCode keycode, unsigned int state)
+{
+  unsigned int cleaned = state & ~(_numlock_physical_mask | LockMask);
+
+  XkbStateRec st;
+  memset (&st, 0, sizeof (st));
+  XkbGetState (dpy, XkbUseCoreKbd, &st);
+
+  unsigned int level = get_level_from_state (dpy, cleaned);
+
+  KeySym ks = XkbKeycodeToKeysym (dpy, keycode, st.group, level);
+  if (ks != NoSymbol && ks != 0)
+    return ks;
+
+  /* Fallback 1: scan levels in current group */
+  for (int lvl = 0; lvl < 8; lvl++)
+    {
+      KeySym t = XkbKeycodeToKeysym (dpy, keycode, st.group, lvl);
+      if (t != NoSymbol && t != 0)
+        return t;
+    }
+
+  /* Fallback 2: scan group 0 */
+  for (int lvl = 0; lvl < 8; lvl++)
+    {
+      KeySym t = XkbKeycodeToKeysym (dpy, keycode, 0, lvl);
+      if (t != NoSymbol && t != 0)
+        return t;
+    }
+
+  return NoSymbol;
+}
+
 static Accessibility_DeviceEvent
 spi_keystroke_from_x_key_event (XKeyEvent *x_key_event)
 {
   Accessibility_DeviceEvent key_event;
   KeySym keysym;
-  const int cbuf_bytes = 20;
-  char cbuf[21];
-  int nbytes;
 
-  nbytes = XLookupString (x_key_event, cbuf, cbuf_bytes, &keysym, NULL);
+  /* Resolve keysym via XKB using current state; honors group and effective modifiers. */
+  keysym = resolve_keysym_with_xkb (spi_get_display (),
+                                    x_key_event->keycode,
+                                    x_key_event->state);
+
   key_event.id = (dbus_int32_t) (keysym);
   key_event.hw_code = (dbus_int16_t) x_key_event->keycode;
   if (((XEvent *) x_key_event)->type == KeyPress)
@@ -820,22 +880,21 @@ spi_keystroke_from_x_key_event (XKeyEvent *x_key_event)
       key_event.event_string = g_strdup ("Right");
       break;
     default:
-      if (nbytes > 0)
-        {
-          gunichar c;
-          cbuf[nbytes] = '\0'; /* OK since length is cbuf_bytes+1 */
-          key_event.event_string = g_strdup (cbuf);
-          c = keysym2ucs (keysym);
-          if (c > 0 && !g_unichar_iscntrl (c))
-            {
-              key_event.is_text = TRUE;
-              /* incorrect for some composed chars? */
-            }
-        }
-      else
-        {
-          key_event.event_string = g_strdup ("");
-        }
+      {
+        gunichar uc = (gunichar) keysym2ucs (keysym);
+        if (uc > 0 && !g_unichar_iscntrl (uc))
+          {
+            char utf8[6];
+            int len = g_unichar_to_utf8 (uc, utf8);
+            utf8[len] = '\0';
+            key_event.event_string = g_strdup (utf8);
+            key_event.is_text = TRUE;
+          }
+        else
+          {
+            key_event.event_string = g_strdup ("");
+          }
+      }
     }
 
   key_event.timestamp = (dbus_uint32_t) x_key_event->time;
